/**
  *  \file game/load-msg.cc
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/load.h"
#include "cpluslib/array.h"
#include "game/msg.h"
#include "game/struct.h"
#include "io/io.h"
#include "io/streams.h"
#include "util/except.h"
#include "util/nls.h"

/* arbitrary limit */
enum { MAX_MSG_SIZE = 16000 };

/** Load version 3.0 outbox (MESSx.DAT).
    \param s Stream to read from */
void
GOutboxReader::loadOutbox(Stream& s)
{
    TInt16 n;

    /* Read count */
    if (!getStructure(s, n) || n.value <= 0)
        return;

    /* Read directory */
    std::size_t header_size = TOutgoingMessageHeader::size * n.value;
    Array<char> header(header_size);
    s.readT(header.ptr(), header_size);

    /* Read content */
    for (int i = 0; i < n.value; ++i) {
        TOutgoingMessageHeader o;
        getStructure(header.ptr() + TOutgoingMessageHeader::size * i, o);
        if (o.address <= 0)
            throw FileFormatException(s, _("Invalid message directory"));
        s.seek(o.address - 1);

        if (o.length < 0 || o.length > MAX_MSG_SIZE)
            throw FileFormatException(s, _("Message too big"));
        if (o.length == 0)
            continue;

        Array<char> data(o.length);
        s.readT(data.ptr(), o.length);

        addMessage(getMessageFromArray(data.ptr(), o.length, false), GPlayerSet(o.to));
    }
}

/** Load version 3.5 outbox (MESS35x.DAT).
    \param s Stream to read from */
void
GOutboxReader::loadOutbox35(Stream& s)
{
    /* file format is
       - word      number of messages
       - 17 bytes  garbage
       - N records: 1 byte    garbage
                    1 byte    "valid" flag
                    12 bytes  receivers
                    word      text length
                    N bytes   text (encrypted) */
    TInt16 n;
    char header[20];

    /* read count */
    if (!getStructure(s, n) || n.value <= 0)
        return;

    /* skip garbage */
    s.readT(header, 17);
    while (n.value-- && s.read(header, 16) == 16) {
        uint16 msgsize = getUint16(header, 14);
        if (msgsize > MAX_MSG_SIZE)
            throw FileFormatException(s, _("Message too big"));
        if (msgsize == 0)
            continue;

        Array<char> msgtext(msgsize);
        s.readT(msgtext.ptr(), msgsize);

        if (header[1] == '1') {
            /* message is valid. Gather receivers */
            GPlayerSet recv;
            for (int i = 1; i <= 12; ++i)
                if (header[1+i] == '1')
                    recv |= i;
            addMessage(getMessageFromArray(msgtext.ptr(), msgsize, false), recv);
        }
    }
}
