/**
  *  \file int/array.cc
  *  \brief Interpreter: Arrays
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/array.h"
#include "int/if/ifutil.h"      // FIXME
#include "int/args.h"
#include "int/interror.h"
#include "int/vmio.h"

/** Maximum total size of an array. We must define an upper limit to
    avoid getting into regions where operator new fails too easily. By
    defining the maximum size as 10001-squared, we allow an array
    indexed by two minefield Ids (the highest-possible Id sequence),
    while having a maximum memory consumption of the IntDataSegment of
    about 400 megs, 1/10 of the 32-bit address space. */
static const uint32_t MAX_ARRAY_ELEMS = 10001UL * 10001UL;

/****************************** IntArrayData *****************************/

/** Constructor. Makes a blank array with no dimensions. */
IntArrayData::IntArrayData()
    : total_size(1),
      dimensions(),
      content()
{ }

/** Destructor. */
IntArrayData::~IntArrayData()
{ }

/** Add a dimension.
    \param n New dimension
    \return true if dimension accepted, false on error */
bool
IntArrayData::addDimension(int32_t n)
{
    /* We must allow dimension 0 to allow empty arrays, so we cannot
       have just one single "n < 0" test. */
    if (n < 0)
        return false;

    if (n > 0 && total_size > 0 && uint32_t(n) > MAX_ARRAY_ELEMS / total_size)
        return false;

    dimensions.push_back(n);
    total_size *= n;
    return true;
}

/** Compute index for this array.
    \param args   [in/out] Argument block containing user-supplied indices
    \param result [out] Linear index
    \retval true Result has been produced
    \retval false One of the indices was null
    \throw IntError One of the indices was out-of-range */
bool
IntArrayData::computeIndex(IntArgBlock& args, uint32_t& result) const
{
    args.checkArgumentCount(dimensions.size());

    uint32_t scale = 1;
    result = 0;
    for (uint32_t i = 0; i < dimensions.size(); ++i) {
        // Read arg. Check for null.
        int32_t me;
        if (!checkIntArg(me, args.getNext()))
            return false;

        // Range check.
        if (me < 0 || uint32_t(me) >= dimensions[i])
            throw IntError::rangeError();

        // Compute index. Horner schema.
        result = result * scale + me;
        scale *= dimensions[i];
    }
    return true;
}


/******************************** IntArray *******************************/

/** Constructor.
    \param data Our array */
IntArray::IntArray(Ptr<IntArrayData> data)
    : data(data)
{ }

// IntIndexableValue:
IntValue*
IntArray::get(IntArgBlock& args)
{
    uint32_t index;
    if (data->computeIndex(args, index)) {
        return IntValue::cloneOf(data->content[index]);
    } else {
        return 0;
    }
}

void
IntArray::set(IntArgBlock& args, IntValue* value)
{
    uint32_t index;
    if (data->computeIndex(args, index)) {
        data->content.set(index, value);
    } else {
        throw IntError::typeError(IntError::tyInteger);
    }
}

IntContext*
IntArray::makeFirstContext()
{
    throw IntError::typeError(IntError::tyIterable);
}

// IntValue:
string_t
IntArray::toString(bool /*readable*/)
{
    return "#<array>";
}

void
IntArray::store(IntTagNode& tag, Stream& /*aux*/)
{
    IntVMSaveContext* vsc = IntVMSaveContext::getCurrentInstance();
    if (vsc != 0) {
        tag.tag   = IntTagNode::Tag_Array;
        tag.value = vsc->addArray(*data);
    } else {
        throw IntError::notSerializable();
    }
}

IntValue*
IntArray::clone()
{
    return new IntArray(data);
}
