/**
  *  \file int/if/baseif.cc
  *  \brief Interpreter Game Interface: Starbase Properties
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/if/baseif.h"
#include "cpluslib/format.h"
#include "game/actions/baseammo.h"
#include "game/actions/basefix.h"
#include "game/actions/baseparts.h"
#include "game/actions/baseship.h"
#include "game/actions/basetech.h"
#include "game/config.h"
#include "game/global.h"
#include "game/planet.h"
#include "game/ship.h"
#include "game/tables.h"
#include "game/undo.h"
#include "int/args.h"
#include "int/compound.h"
#include "int/exec.h"
#include "int/if/ifutil.h"
#include "int/interror.h"
#include "int/value.h"
#include "game/planetform.h"
#include "game/planettransfer.h"
#include "game/shiptransfer.h"
#include "game/actions/preconditions.h"

namespace {
    /** Starbase array property. This handles a starbase's array properties, Storage.Xxx(). */
    class IntBaseArrayProperty : public IntIndexableValue {
        const GPlanet& p;
        const IntBaseProperty ibp;

        IntValue* performArrayReference(mp16_t (GPlanet::*func)(int) const, int32_t limit, int32_t arg, bool hull);
     public:
        IntBaseArrayProperty(const GPlanet& p, IntBaseProperty ibp);

        // IntIndexableValue:
        IntValue* get(IntArgBlock& args);
        void set(IntArgBlock& args, IntValue* value);
        IntContext* makeFirstContext();

        // IntValue:
        IntBaseArrayProperty* clone();
        string_t toString(bool readable);
        void store(IntTagNode& sv, Stream& aux);
    };
}

/** Constructor.
    \param p Starbase
    \param ibp Property to implement */
IntBaseArrayProperty::IntBaseArrayProperty(const GPlanet& p, IntBaseProperty ibp)
    : p(p), ibp(ibp)
{ }

/** Perform array reference. This implements the special abilities of starbase array properties:
    - index 0 counts all items
    - index 1..limit returns that item's count
    - other values yield empty (not an error!)
    \param func GPlanet function to access the property in raw form (e.g. getBaseEngineStore)
    \param limit Maximum index (e.g. NUM_ENGINES)
    \param arg User-specified argument
    \return resulting value */
IntValue*
IntBaseArrayProperty::performArrayReference(mp16_t (GPlanet::*func)(int) const, int32_t limit, int32_t arg, bool hull)
{
    if (arg == 0) {
        int32_t sum = 0;
        for (int i = 1; i <= limit; ++i) {
            mp16_t val = (p.*func)(i);
            if (!val.isKnown()) {
                return 0;
            }
            sum += val;
        }
        return makeIntValue(sum);
    } else if (hull) {
        if (arg < 0 || arg > getNumHulls())
            return 0;
        if (int n = getTruehullSlot(p.getOwner(), arg))
            return makeIntFromMaybe((p.*func)(n));
        else
            return makeIntValue(0);
    } else if (arg > 0 && arg <= limit) {
        return makeIntFromMaybe((p.*func)(arg));
    } else {
        return 0;
    }
}

IntValue*
IntBaseArrayProperty::get(IntArgBlock& args)
{
    int32_t arg;
    args.checkArgumentCount(1);
    if (!checkIntArg(arg, args.getNext()))
        return 0;

    switch (ibp) {
     case ibpEngineStorage:
        return performArrayReference(&GPlanet::getBaseEngineStore, NUM_ENGINES, arg, false);
     case ibpHullStorage:
        return performArrayReference(&GPlanet::getBaseHullStoreSlot, NUM_HULLS_PER_PLAYER, arg, true);
     case ibpBeamStorage:
        return performArrayReference(&GPlanet::getBaseBeamStore, NUM_BEAMS, arg, false);
     case ibpLauncherStorage:
        return performArrayReference(&GPlanet::getBaseLauncherStore, NUM_TORPS, arg, false);
     case ibpAmmoStorage:
        return performArrayReference(&GPlanet::getBaseAmmoStore, NUM_TORPS+1, arg, false);
     default:
        return 0;
    }
}

void
IntBaseArrayProperty::set(IntArgBlock& /*args*/, IntValue* /*value*/)
{
    throw IntError::notAssignable();
}

IntContext*
IntBaseArrayProperty::makeFirstContext()
{
    throw IntError::typeError(IntError::tyIterable);
}

IntBaseArrayProperty*
IntBaseArrayProperty::clone()
{
    return new IntBaseArrayProperty(p, ibp);
}

string_t
IntBaseArrayProperty::toString(bool /*readable*/)
{
    return "#<array>";
}

void
IntBaseArrayProperty::store(IntTagNode& /*sv*/, Stream& /*aux*/)
{
    throw IntError::notSerializable();
}


/** Get starbase property.
    \param pl Starbase
    \param ibp Starbase property to query */
IntValue*
getBaseProperty(const GPlanet& pl, GUniverse& univ, IntBaseProperty ibp)
{
    if (!pl.isPlayed() || !pl.hasBase())
        return 0;

    switch (ibp) {
     case ibpBaseDamage:
        return makeIntFromMaybe(pl.getBaseDamage());
     case ibpBaseDefense:
        return makeIntValue(pl.getStructures(str_BaseDefense));
     case ibpBaseDefenseMax:
        return makeIntValue(getMaxStructures(pl, str_BaseDefense));
     case ibpBaseFighters:
        return makeIntFromMaybe(pl.getCargo(el_Fighters));
     case ibpBaseFightersMax:
        return makeIntValue(config.MaximumFightersOnBase(pl.getOwner()));
     case ibpBeamTech:
        return makeIntFromMaybe(pl.getBaseTechLevel(tl_Beam));
     case ibpBuildBeam:
        return makeIntFromMaybe(mp16_t(pl.getBaseBuildOrder().beam_type));
     case ibpBuildBeamCount:
        return makeIntFromMaybe(mp16_t(pl.getBaseBuildOrder().beam_count));
     case ibpBuildEngine:
        return makeIntValue(mp16_t(pl.getBaseBuildOrder().engine));
     case ibpBuildFlag:
        return makeBoolValue(mp16_t(pl.getBaseBuildHull() != 0));
     case ibpBuildHull:
        if (pl.getBaseBuildHull().isKnown() && pl.getBaseBuildHull() > 0)
            return makeIntValue(pl.getBaseBuildHull());
        else
            return 0;
     case ibpBuildHullName:
        if (pl.getBaseBuildHull().isKnown()) {
            if (int n = pl.getBaseBuildHull()) {
                return makeStringValue(::getHull(n).getName());
            }
        }
        return 0;
     case ibpBuildHullShort:
        if (pl.getBaseBuildHull().isKnown()) {
            if (int n = pl.getBaseBuildHull()) {
                return makeStringValue(::getHull(n).getShortName());
            }
        }
        return 0;
     case ibpBuildQueuePos:
        return makeIntFromMaybe(pl.getBaseQueuePosition());
     case ibpBuildTorp:
        return makeIntFromMaybe(mp16_t(pl.getBaseBuildOrder().torp_type));
     case ibpBuildTorpCount:
        return makeIntFromMaybe(mp16_t(pl.getBaseBuildOrder().torp_count));
     case ibpEngineTech:
        return makeIntFromMaybe(pl.getBaseTechLevel(tl_Engine));
     case ibpHullTech:
        return makeIntFromMaybe(pl.getBaseTechLevel(tl_Hull));
     case ibpMission:
        return makeIntFromMaybe(pl.getBaseMission());
     case ibpMissionName:
        {
            mp16_t msn = pl.getBaseMission();
            if (msn.isKnown()) {
                return makeStringValue(getBaseMissionName(msn));
            } else {
                return 0;
            }
        }
     case ibpShipyardAction:
        if (pl.getBaseShipyardOrder().isKnown()) {
            switch (pl.getBaseShipyardOrder()) {
             case sya_Fix:     return makeStringValue("Fix");
             case sya_Recycle: return makeStringValue("Recycle");
            }
        }
        return 0;
     case ibpShipyardId:
        return makeIntFromMaybe(pl.getBaseShipyardId());
     case ibpShipyardName:
        if (pl.getBaseShipyardId().isKnown()) {
            if (int sid = pl.getBaseShipyardId()) {
                return makeStringValue(univ.getShip(sid).getName(GObject::PlainName));
            }
        }
        return 0;
     case ibpShipyardStr:
        if (pl.getBaseShipyardOrder().isKnown() && pl.getBaseShipyardId().isKnown()) {
            switch (pl.getBaseShipyardOrder()) {
             case sya_Fix:     return makeStringValue(format("Fix %s", univ.getShip(pl.getBaseShipyardId()).getName(GObject::PlainName)));
             case sya_Recycle: return makeStringValue(format("Recycle %s", univ.getShip(pl.getBaseShipyardId()).getName(GObject::PlainName)));
            }
        }
        return 0;
     case ibpTorpedoTech:
        return makeIntFromMaybe(pl.getBaseTechLevel(tl_Torpedo));

     case ibpEngineStorage:
     case ibpHullStorage:
     case ibpBeamStorage:
     case ibpLauncherStorage:
     case ibpAmmoStorage:
        return new IntBaseArrayProperty(pl, ibp);
    }
    return 0;
}

void
setBaseProperty(GPlanet& pl, IntBaseProperty ibp, IntValue* value)
{
    if (!pl.hasBase() || !pl.isPlayed())
        throw IntError::notAssignable();

    int32_t iv;
    switch (ibp) {
     case ibpMission:
        // FIXME: we don't have a constant for '6'.
        if (checkIntArg(iv, value, 0, 6)) {
            pl.setBaseMission(mp16_t(iv));
        }
        break;
     default:
        throw IntError::notAssignable();
    }
}

static void
setBaseOrder(GUniverse& univ, TShipyardAction order, GPlanet& pl, int sid)
{
    // Verify
    if (sid == 0) {
        // We want to cancel the order. So all we need is a base.
        mustHavePlayedBase(pl);
        order = sya_None;
    } else {
        // Verify order.
        if (GError e = checkBaseAction(order, univ, sid, pl.getId()))
            throw e;
    }

    // Do it
    pl.setBaseShipyardOrder(mp16_t(order), mp16_t(sid));
}

/** Backend for BuildXXX component build commands.
    \param pl   Planet to work on
    \param exc  Execution context (required to assign BUILD.REMAINDER)
    \param area Kind of component to build
    \param type Type of component to build (index, 1..NUM_HULLS_PER_PLAYER/NUM_ENGINES/...)
    \param amount Amount to build or scrap
    \param flag true to permit partial builds */
static void
doBaseBuildComponents(GPlanet& pl, IntExecutionContext& exc, TTechLevel area, int type, int amount, bool flag)
{
    // Do it (checks preconditions)
    GStarbaseBuildPartsAction action(pl);
    action.addUndoInformation(getDisplayedTurn().getPreviousUniverse().getPlanet(pl.getId()));
    int32_t did = action.addParts(area, type, amount, flag);

    // Now (try to) do it
    GMultiTransaction mt;
    GPlanetTransfer& tx           = mt.addNewSubtransaction(new GPlanetTransfer(pl));
    GStarbaseBuildTransaction& ta = mt.addNewSubtransaction(new GStarbaseBuildTransaction(action, tx));

    if (flag != 0) {
        // Try to back out if we have too little cash, then commit (this will throw if there is a different error).
        while (did > 0 && !ta.isValidCost() && action.addParts(area, type, -1, false))
            --did;
        mt.doCommit();

        // Tell user about remainder
        IntIntIntValue iv(amount - did);
        exc.setVariable("BUILD.REMAINDER", &iv);
    } else {
        // We are not permitted to do a partial build.
        if (did != amount) {
            mt.doCancel();
            throw GError(GError::eNoResource, GError::eNoResource);
        } else {
            mt.doCommit();
        }
    }
}

/** Backend for BuildXXX ammo build commands.
    \param pl   Planet to work on
    \param univ Universe to work in
    \param exc  Execution context (required to assign BUILD.REMAINDER)
    \param type Type of component to build
    \param amount Amount to build or scrap
    \param flag true to permit partial builds
    \param sid  Target ship Id */
static void
doBaseBuildAmmo(GPlanet& pl, GUniverse& univ, IntExecutionContext& exc, int type, int amount, bool flag, int32_t sid)
{
    // Do we have a base?
    mustHavePlayedBase(pl);

    // Do we have a valid ship Id?
    if (sid < 0 || sid > NUM_SHIPS)
        throw IntError::rangeError();
    if (sid > 0) {
        const GShip& sh = univ.getShip(sid);
        if (!sh.isPlayed() || sh.getPos() != pl.getPos())
            throw GError(GError::ePos, GError::ePos);
    }

    // Build objects.
    GMultiTransaction mt;
    GPlanetTransfer& payAmmo       = mt.addNewSubtransaction(new GPlanetTransfer(pl));
    GPlanetTransfer& payTech       = mt.addNewSubtransaction(new GPlanetTransfer(pl));
    GCargoContainer& receiver      = (sid > 0
                                      ? static_cast<GCargoContainer&>(mt.addNewSubtransaction(new GShipTransfer(univ.getShip(sid))))
                                      : static_cast<GCargoContainer&>(mt.addNewSubtransaction(new GPlanetTransfer(pl))));
    GStarbaseBuildAmmoAction& ammo = mt.addNewSubtransaction(new GStarbaseBuildAmmoAction(payAmmo, receiver));
    GStarbaseTechUpgradeAction tech(pl);
    mt.addNewSubtransaction(new GStarbaseBuildTransaction(tech, payTech));
    // Phew.

    // If we're going to build something, so make sure we have the tech
    if (amount > 0) {
        int needed = ammo.getTechUpgradeNeeded(type);
        if (needed) {
            if (!tech.setTech(tl_Torpedo, needed))
                throw GError(GError::ePerm, GError::ePerm);
            if (GError e = mt.getError())
                throw e;
        }
    }

    // Build
    int32_t n = ammo.limitAmount(type, amount);
    if (flag) {
        ammo.setAmount(type, ammo.getAmount(type) + n);
        mt.doCommit();

        IntIntIntValue iv(amount - n);
        exc.setVariable("BUILD.REMAINDER", &iv);
    } else {
        if (n != amount)
            throw GError(GError::eNoResource, GError::eNoResource);
        ammo.setAmount(type, ammo.getAmount(type) + n);
        mt.doCommit();
    }
}

/*+ BuildFighters <amount>[, <flagAndShipId>]

    Attempts to build <amount> fighters. The amount can be negative to
    scrap fighters. The <flagAndShipId> can be "N" to permit partial
    builds. If not all of the requested amount can be built, the
    command will report the amount not built in the variable
    BUILD.REMAINDER instead of failing.

    <flagAndShipId> can also contain a ship Id, to place the
    newly-built fighters on that ship. */
void
IFBaseBuildFighters(GObject& obj, GUniverse& univ, IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(1, 2);

    int32_t amount;
    if (!checkIntArg(amount, args.getNext(), -10000, 10000))
        return;

    int32_t flag = 0;
    int32_t sid = 0;
    checkFlagArg(flag, &sid, args.getNext(), "N");

    doBaseBuildAmmo(static_cast<GPlanet&>(obj), univ, exc, GStarbaseBuildAmmoAction::max_index, amount, flag, sid);
}

/*+ BuildEngines <type>, <amount>[, <flag>]

    Attempts to build <amount> engines of the given <type>. The amount
    can be negative to scrap engines. The tech levels is automatically
    raised as necessary. The <flag> can be "N" to permit partial
    builds. If not all of the requested amount can be built, the
    command will report the amount not built in the variable
    BUILD.REMAINDER instead of failing. */
void
IFBaseBuildEngines(GObject& obj, GUniverse& /*univ*/, IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(2, 3);

    int32_t type;
    if (!checkIntArg(type, args.getNext(), 1, NUM_ENGINES))
        return;

    int32_t amount;
    if (!checkIntArg(amount, args.getNext(), -10000, +10000))
        return;

    int32_t flag = 0;
    checkFlagArg(flag, 0, args.getNext(), "N");

    doBaseBuildComponents(static_cast<GPlanet&>(obj), exc, tl_Engine, type, amount, flag);
}

/*+ BuildHulls <type>, <amount>[, <flag>]

    Attempts to build <amount> hulls of the given <type>. The amount
    can be negative to scrap hulls. The tech levels is automatically
    raised as necessary. The <flag> can be "N" to permit partial
    builds. If not all of the requested amount can be built, the
    command will report the amount not built in the variable
    BUILD.REMAINDER instead of failing.

    The <type> is a hull Id. You can not build all hulls. */
void
IFBaseBuildHulls(GObject& obj, GUniverse& /*univ*/, IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(2, 3);

    int32_t type;
    if (!checkIntArg(type, args.getNext(), 1, getNumHulls()))
        return;

    int32_t amount;
    if (!checkIntArg(amount, args.getNext(), -10000, +10000))
        return;

    int32_t flag = 0;
    checkFlagArg(flag, 0, args.getNext(), "N");

    // Can we build this hull?
    int32_t slot = getTruehullSlot(obj.getOwner(), type);
    if (!slot) {
        if (amount != 0)
            throw GError(GError::ePerm, GError::ePerm);
    } else {
        doBaseBuildComponents(static_cast<GPlanet&>(obj), exc, tl_Hull, slot, amount, flag);
    }
}

/*+ BuildLaunchers <type>, <amount>[, <flag>]

    Attempts to build <amount> torpedo launchers of the given <type>.
    The amount can be negative to scrap launchers. The tech levels is
    automatically raised as necessary. The <flag> can be "N" to permit
    partial builds. If not all of the requested amount can be built,
    the command will report the amount not built in the variable
    BUILD.REMAINDER instead of failing. */
void
IFBaseBuildLaunchers(GObject& obj, GUniverse& /*univ*/, IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(2, 3);

    int32_t type;
    if (!checkIntArg(type, args.getNext(), 1, NUM_TORPS))
        return;

    int32_t amount;
    if (!checkIntArg(amount, args.getNext(), -10000, +10000))
        return;

    int32_t flag = 0;
    checkFlagArg(flag, 0, args.getNext(), "N");

    doBaseBuildComponents(static_cast<GPlanet&>(obj), exc, tl_Torpedo, type, amount, flag);
}

/*+ BuildBeams <type>, <amount>[, <flag>]

    Attempts to build <amount> beams of the given <type>. The amount
    can be negative to scrap beams. The tech levels is automatically
    raised as necessary. The <flag> can be "N" to permit partial
    builds. If not all of the requested amount can be built, the
    command will report the amount not built in the variable
    BUILD.REMAINDER instead of failing. */
void
IFBaseBuildBeams(GObject& obj, GUniverse& /*univ*/, IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(2, 3);

    int32_t type;
    if (!checkIntArg(type, args.getNext(), 1, NUM_BEAMS))
        return;

    int32_t amount;
    if (!checkIntArg(amount, args.getNext(), -10000, +10000))
        return;

    int32_t flag = 0;
    checkFlagArg(flag, 0, args.getNext(), "N");

    doBaseBuildComponents(static_cast<GPlanet&>(obj), exc, tl_Beam, type, amount, flag);
}

/*+ BuildTorps <type>, <amount>[, <flagAndShipId>]

    Attempts to build <amount> torpedoes of the given type. The amount
    can be negative to scrap torpedoes. The <flagAndShipId> can be "N"
    to permit partial builds. If not all of the requested amount can
    be built, the command will report the amount not built in the
    variable BUILD.REMAINDER instead of failing.

    <flagAndShipId> can also contain a ship Id, to place the
    newly-built torpedoes on that ship. */
void
IFBaseBuildTorps(GObject& obj, GUniverse& univ, IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(2, 3);

    int32_t type;
    if (!checkIntArg(type, args.getNext(), 1, NUM_TORPS))
        return;

    int32_t amount;
    if (!checkIntArg(amount, args.getNext(), -10000, 10000))
        return;

    int32_t flag = 0;
    int32_t sid = 0;
    checkFlagArg(flag, &sid, args.getNext(), "N");

    doBaseBuildAmmo(static_cast<GPlanet&>(obj), univ, exc, type, amount, flag, sid);
}

/** Parse build ship command.
    \param args [in] arguments
    \param o [out] build order
    \retval true correctly parsed
    \retval false null argument encountered
    \throws GError or IntError on error */
bool
parseBuildShipCommand(IntArgBlock& args, TBuildOrder& o)
{
    args.checkArgumentCount(1, 6);

    // Mandatory arg
    int32_t hull;
    if (!checkIntArg(hull, args.getNext(), 0, getNumHulls())) {
        return false;
    }

    // Optional args
    int32_t engine = 0, beam = 0, beamcount = -1, torp = 0, torpcount = -1;
    checkIntArg(engine, args.getNext(), 1, NUM_ENGINES);
    checkIntArg(beam, args.getNext(), 0, NUM_BEAMS);
    checkIntArg(beamcount, args.getNext());
    checkIntArg(torp, args.getNext(), 0, NUM_BEAMS);
    checkIntArg(torpcount, args.getNext());

    // Check mandatory arg
    o.hull = hull;
    if (hull == 0) {
        // Canceling a build
        return true;
    }

    // This is a ship build order. Validate remaining args
    const GHull& h = getHull(hull);

    // Engine
    if (engine <= 0)
        throw IntError::rangeError();
    o.engine = engine;

    // Beams
    if (beamcount == -1)
        beamcount = h.getMaxBeams();
    if (beam == 0 || beamcount == 0) {
        o.beam_type = o.beam_count = 0;
    } else if (beamcount < 0 || beamcount > h.getMaxBeams()) {
        throw IntError::rangeError();
    } else {
        o.beam_type  = beam;
        o.beam_count = beamcount;
    }

    // Torps
    if (torpcount == -1)
        torpcount = h.getMaxLaunchers();
    if (torp == 0 || torpcount == 0) {
        o.torp_type = o.torp_count = 0;
    } else if (torpcount < 0 || torpcount > h.getMaxLaunchers()) {
        throw IntError::rangeError();
    } else {
        o.torp_type  = torp;
        o.torp_count = torpcount;
    }
    return true;
}

/*+ BuildShip <hull>[, <engine>, <beamtype>, <beamcount>, <torptype>, <torpcount>]

    Set the ship build order. If <hull> is zero, cancel a pending
    order. Otherwise, build a ship. In this case, <engine> must be
    specified, and the others should better be specified as well to
    avoid building a ship without weapons.

    Required tech levels and parts are bought automatically. */
void
IFBaseBuildShip(GObject& obj, GUniverse& /*univ*/, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    // Parse args
    TBuildOrder o;
    if (!parseBuildShipCommand(args, o))
        return;

    // Get planet
    GPlanet& pl = static_cast<GPlanet&>(obj);
    mustHavePlayedBase(pl);

    // Check for cancellation
    if (o.hull == 0) {
        pl.setBaseBuildOrder(o);
        return;
    }

    // Make a transaction and fire it
    GStarbaseBuildShipAction action(pl);
    action.setUsePartsFromStorage(true);
    action.setUseTechUpgrade(true);
    action.setBuildOrder(o);

    GMultiTransaction mt;
    GPlanetTransfer& tx = mt.addNewSubtransaction(new GPlanetTransfer(pl));
    mt.addNewSubtransaction(new GStarbaseBuildTransaction(action, tx));
    mt.doCommit();
    pl.setBaseBuildOrder(action.getBuildOrder()); // FIXME: why is this separate?
}

/*+ FixShip <sid>

    Fix (repair) a ship. The <sid> is a ship Id, or 0 to cancel a
    pending shipyard order. */
void
IFBaseFixShip(GObject& obj, GUniverse& univ, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    int32_t sid;
    args.checkArgumentCount(1);
    if (checkIntArg(sid, args.getNext(), 0, NUM_SHIPS))
        setBaseOrder(univ, sya_Fix, static_cast<GPlanet&>(obj), sid);
}

/*+ RecycleShip <sid>

    Recycle a ship. The <sid> is a ship Id, or 0 to cancel a pending
    shipyard order. */
void
IFBaseRecycleShip(GObject& obj, GUniverse& univ, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    int32_t sid;
    args.checkArgumentCount(1);
    if (checkIntArg(sid, args.getNext(), 0, NUM_SHIPS))
        setBaseOrder(univ, sya_Recycle, static_cast<GPlanet&>(obj), sid);
}

/*+ SetMission <number>

    Set starbase mission. */
void
IFBaseSetMission(GObject& obj, GUniverse& /*univ*/, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(1);
    setBaseProperty(static_cast<GPlanet&>(obj), ibpMission, args.getNext());
}

/*+ SetBaseTech <area>, <level>

    Set starbase tech level. <area> is 1 for engines, 2 for hulls, 3
    for beams, 4 for torpedoes. <level> is the new tech level. */
void
IFBaseSetTech(GObject& obj, GUniverse& /*univ*/, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(2);

    // Fetch arguments
    int32_t area, tech;
    if (!checkIntArg(area, args.getNext(), 1, 4))
        return;
    if (!checkIntArg(tech, args.getNext(), 1, 10))
        return;
    --area;

    // Get planet
    GPlanet& pl = static_cast<GPlanet&>(obj);

    // Create tech upgrade action (checks preconditions)
    GStarbaseTechUpgradeAction action(pl);
    GUndoInformation undo;
    undo.computeInformation(getDisplayedTurn(), pl.getId());
    action.setUndoInformation(undo);

    // Do the rules permit this?
    if (!action.setTech(area, tech))
        throw GError(GError::ePerm, GError::ePerm);

    // Now (try to) do it
    GMultiTransaction mt;
    GPlanetTransfer& tx = mt.addNewSubtransaction(new GPlanetTransfer(pl));
    mt.addNewSubtransaction(new GStarbaseBuildTransaction(action, tx));
    mt.doCommit();
}
